// dbg_wrapper.hpp 
//
//////////////////////////////////////////////////////////////////////

// Debug Utility library
//
// Copyright (C) 2003 John Torjo (john@torjo.com)
//
// Permission to copy, use, sell and distribute this software is granted
// provided this copyright notice appears in all copies.
// Permission to modify the code and to distribute modified code is granted
// provided this copyright notice appears in all copies, and a notice
// that the code was modified is included with the copyright notice.
//
// This software is provided "as is" without express or implied warranty,
// and with no claim as to its suitability for any purpose.
 
// See http://www.boost.org for updates, documentation, and revision history.

#if !defined(BOOST_DBG_WRAPPER_HPP)
#define BOOST_DBG_WRAPPER_HPP

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include <iosfwd>
#include <string>
#include <sstream>

/*
    tested with:
    VC6
    gcc 3.2
    como 4.3.2 (compiles, fails to link)
*/

#undef BOOST_DBG_WRAPPER_VC6
#if defined( _MSC_VER) 
#if _MSC_VER < 1300
#define BOOST_DBG_WRAPPER_VC6
#endif
#endif



namespace boost { namespace dbgutil {



/*
    how are we wrapping a handle-like type?
*/
typedef enum {
    // behaves like the underlying handle-like type
    //
    // keeps a user-friendly string in itself, 
    // so that at any time you can see the user-friendly string
    // (very useful for debugging)
    wrapper_debug,


    // behaves like the underlying handle-like type
    // when written to stream, writes only time_t value (no user-friendly string)
    // when read from stream, ignores user-friendly msg. if necessary
    wrapper_release,

    wrapper_default = 
#ifndef NDEBUG
    // debug
    wrapper_debug
#else
    // release
    wrapper_release
#endif

} dbg_wrapper_type;


// forward declaration
template< class handle_traits, dbg_wrapper_type wrapper_type, class char_type>
struct dbg_wrapper;


// traits for dealing with streams
//

/*
    top class for dbg_wrapper_traits

    Specialize this in case you're not satisfied with the prefix().

    When printing *any* handle object, it should not contain
    the prefix(). This way, when reading a dbg_wrapper<handle> value
    from a stream, by reading a character we know if it contains
    details or not.

  
    For instance, time_t values are composed of digits.
    Therefore, any letter can be an acceptable prefix.
*/
template< class handle_traits>
struct dbg_wrapper_traits_top {
    static char prefix() { return 'D'; }
};


// forward declaration
template< class handle_traits, class char_type>
struct dbg_wrapper_traits;


/* 
    don't specialize this.

    Specialize dbg_wrapper_traits instead!
*/
template< class handle_traits, class char_type>
struct dbg_wrapper_traits_base : public dbg_wrapper_traits_top< handle_traits> {

    typedef dbg_wrapper_traits< handle_traits, char_type> traits;
    typedef typename handle_traits::raw_type handle;

    /*
        the type that contains information about the handle,
        when debugging 
        (in dbg_wrapper<handle, wrapper_debug> class)
    
        the 'info_type' must be copy-constructible
    */
    typedef std::basic_string< char_type> info_type;


    /* 
        writes the value & details about it
        ('info' contains the debug information about the 'val')
    */
    template< class other_char_type, class traits_type>
    static void write_val_and_details( 
            handle val, 
            const info_type & info, 
            std::basic_ostream< other_char_type, traits_type> & out) {
        out << prefix();
        traits::write_val( val, out);
        out << ' ' << info;
    }


    /* 
        writes ONLY the value (no details about it)
    */
    template< class other_char_type, class traits_type>
    static void write_val( handle val, std::basic_ostream< other_char_type, traits_type> & out) {
        out << val;

    }

    /*
        reads the details of the given 'val'
    */
    static void read_details( handle val, info_type & info) {
        std::basic_ostringstream< char_type> out;
        traits::write_details( val, out);
        info = out.str();
    }
};


/*
    specialize this for your handle class, providing
    the shown below functions.

    It's recommended that you derive from dbg_wrapper_traits_base<>,
    or you'll have to provide those functions as well
    (see dbg_wrapper_traits_base<> above)

    See dbg_wrapper_time_t.hpp for examples of how to do this.
*/
template< class handle_traits, class char_type>
struct dbg_wrapper_traits : public dbg_wrapper_traits_base< handle_traits, char_type> {

    typedef typename handle_traits::raw_type handle;
    /* 
        writes ONLY details about the value
    */
    template< class other_char_type, class traits_type>
    static void write_details( 
            handle val, 
            std::basic_ostream< other_char_type, traits_type> & out);

    /*
        read the value (ignoring the details, if any)
    */
    template< class other_char_type, class traits_type>
    static void read_val( 
            handle &val, 
            std::basic_istream< other_char_type, traits_type> & in);

};




namespace detail {
    struct dbg_wrapper_base {
        static bool & write_details_val() {
            static bool val =
    #ifndef NDEBUG
                // debug
                true
    #else
                // release
                false
    #endif
                ;
                return val;
        }
    }; 

    // note: we derive the dbg_wrapper from representatin<wrapper>::wrapper<handle>
    //
    // (done with inner class, so that it works for compilers
    //  that don't support partial specialization)
    template< dbg_wrapper_type wrapper_type>
    struct representation 
#if !defined( BOOST_DBG_WRAPPER_VC6)
    ;
#else
    // needed because of VC6 bug
    {
        template< class handle_traits, class char_type>
            struct wrapper {};
    };
#endif

    template<>
    struct representation< wrapper_debug> {
        template< class handle_traits, class char_type>
        struct wrapper {
            typedef typename ::boost::dbgutil::dbg_wrapper< handle_traits, wrapper_debug, char_type> derived;
            typedef dbg_wrapper_traits< handle_traits, char_type> traits;
            typedef typename traits::info_type info_type;
            typedef typename handle_traits::raw_type handle;

            void refresh( handle val) {
                // update details
                traits::read_details( val, m_info);
            }

            template< class other_char_type, class traits_type>
            void read( handle & val, std::basic_istream< other_char_type, traits_type> & in) {
                traits::read_val( val);
                // update details as well
                refresh();
            }

            template< class other_char_type, class traits_type>
            void write( handle val, std::basic_ostream< other_char_type, traits_type> & out) const {
                if ( derived::write_details() ) 
                    traits::write_val_and_details( val, m_info, out); 
                else 
                    // no details
                    traits::write_val( val, out);
            }

            // user friendly msg
            info_type m_info;
        };
    };

    template<>
    struct representation< wrapper_release> {
        template< class handle_traits, class char_type>
        struct wrapper {
            typedef typename ::boost::dbgutil::dbg_wrapper< handle_traits, wrapper_release, char_type> derived;
            typedef dbg_wrapper_traits< handle_traits, char_type> traits;
            typedef typename traits::info_type info_type;
            typedef typename handle_traits::raw_type handle;

            void refresh( handle) {
                // nothing to refresh - we're on release
            }

            template< class other_char_type, class traits_type>
            void read( handle & val, std::basic_istream< other_char_type, traits_type> & in) {
                traits::read_val( val, in);
            }

            template< class other_char_type, class traits_type>
            void write( handle val, std::basic_ostream< other_char_type, traits_type> & out) const {
                if ( derived::write_details() ) {
                    out << traits::prefix();
                    traits::write_val( val, out); 
                    out << ' ';
                    traits::write_details( val, out);
                }
                else 
                    // no details
                    traits::write_val( val, out);
            }
        };
    };


} // detail



/*
    It allows wrapping a handle-like type for 
    debugging/ testing purposes.

    The reason behind all this is the fact that handles are pretty
    opaque data, so dealing with them is pretty hard.
    Bugs are very hard to track/ solve.

    Therefore, when using handles, you could use some more information
    about them (depending on their type).

    For instance, for a time_t value, it would be very useful
    to have the time information as a string as well.

    Example: for time_t value '1061459206', the string information
    could look like: '2003_aug_21___12:46:46' (for GMT+2).


    Providing more information is done in two ways, while keeping it efficient:
    1. debugging information
    2. writing to/reading from streams

    
    1. When using dbg_wrapper, you can specify what type of debugging info. you want:
    (the 'dbg_wrapper_type' template parameter):
    - wrapper_debug: 
      each 'dbg_wrapper<>' instance contains debug information,
      in an extra variable: m_info.                    
    - wrapper_release
      each 'dbg_wrapper<>' instance DOES NOT contain debug info.
      It behaves like the underlying handle type
      (except for reading/writing to streams, see below)
    - wrapper_default (this is the default)
      defaults to wrapper_debug in debug mode (NDEBUG not defined)
               to wrapper_release in release mode (NDEBUG defined)

    2. writing to/reading from streams
    When writing to streams, we can:
   a) write_details() = false;
      write it just like the underlying handle type
      (for instance, for a time_t, write the time_t value)
   b) write_details() = true;
      write details about it also (for debugging/testing purposes)
      This is very useful when you export/import information into
      files, but again reading opaque handle values does not help
      you very much. Therefore, writing some more info could about 
      handles could be of good use.

    Writing details is done using the following convension:
    '<prefix><handle-value> <details about the handle>'

    Example, for time_t, 
    writing a): '1061459206'
            b): 'D1061459206 2003_aug_21___12:46:46' (for GMT+2)

    
    Reading from streams recognizes BOTH formats. Therefore you
    can interoperate with other apps., different versions of yourself, etc.

    Or, for instance, when you decide your application is stable enough
    and does not need this extra info., you can turn it off, but
    still be able to import/use existing files (streams).
    
    Or, in a client/server application, the client can write with
    write_details() = false and the server with write_details() = true
    (or viceversa).



    For more details, check out 
    dbg_wrapper_traits<>, dbg_wrapper_traits_base<>, dbg_wrapper_traits_top<> classes.



  Efficiency.
      If you worry about efficiency, know that dbg_wrapper<> is as
      efficient as the underlying type in release mode.
      It is slightly less efficient when writing to/reading from
      streams (since it has to recognize both formats, when
      write_details() = false and write_details() = true )
    


 Note about handle_traits

    It allows you to specify the handle class.

    Using the type directly (for instance, time_t, directly)
    won't work, since we might have multiple handle types 
    that map to the same underlying type.

    For instance, HWND and HINTERNET are different types of handles,
    but they both map to 'void *'.

    The handle_traits has to contain a definition of raw_type.
    Example:

    struct time_t_traits {
        typedef time_t raw_type;
    };
  
*/
template< 
            class handle_traits, 
            dbg_wrapper_type wrapper_type = wrapper_default, 
            class char_type = char>
struct dbg_wrapper 
    : private detail::dbg_wrapper_base, 
      // benefit from the Empty Class Optimization, if possible
      private detail::representation< wrapper_type>::wrapper< handle_traits, char_type> {

    typedef typename detail::representation< wrapper_type>::template wrapper< handle_traits, char_type> base;
    typedef typename handle_traits::raw_type handle;

    dbg_wrapper( handle val = handle() ) : m_val( val) {
        base::refresh( m_val);
    }

    operator handle() const {
        return m_val;
    }

    dbg_wrapper& operator=( handle val) {
        m_val = val;
        base::refresh( m_val);
        return *this;
    }

    dbg_wrapper& operator=( const dbg_wrapper & val) {
        m_val = val.m_val;
        base::refresh( m_val);
        return *this;
    }

    // helpers for when you know 
    // EXACTLY how you want to write a time_t_wrapper value
    handle raw() const {
        return m_val;
    }
    // FIXME - will be implemented later ;)
    //detail::stream_writer< handle> details() const {
      //  return detail::stream_writer< handle>( m_val);
    //}

    static bool & write_details() {
        return write_details_val();
    }

    
    // ...made public due to VC6 bug
    template< class other_char_type, class traits_type>
    void do_write( std::basic_ostream< other_char_type, traits_type> &out) const {
        base::write( m_val, out);
    }

    template< class other_char_type, class traits_type>
    void do_read( std::basic_istream< other_char_type, traits_type> & in) {
        base::read( m_val, in);
    }


private:
    handle m_val;
};


}} // namespaces



// for VC6, we need a workaround in order to allow writing to/reading from
// streams
#ifndef BOOST_DBG_WRAPPER_VC6

template< 
            class char_type, 
            class traits_type, 
            class handle_traits, 
            boost::dbgutil::dbg_wrapper_type wrapper_type, 
            class other_char_type>
std::basic_ostream< char_type, traits_type> & operator<<( 
        std::basic_ostream< char_type, traits_type> & out, 
        const boost::dbgutil::dbg_wrapper< handle_traits, wrapper_type, other_char_type> & val) {
    val.do_write( out);
    return out;
}


template< 
            class char_type, 
            class traits_type, 
            class handle_traits, 
            boost::dbgutil::dbg_wrapper_type wrapper_type, 
            class other_char_type>
std::basic_istream< char_type, traits_type> & operator>>( 
        std::basic_istream< char_type, traits_type> & in, 
        boost::dbgutil::dbg_wrapper< handle_traits, wrapper_type, other_char_type> & val) {
    val.do_read( in);
    return in;
}

#else 

// workaround for VC6
template< 
            class handle_traits, 
            boost::dbgutil::dbg_wrapper_type wrapper_type, 
            class other_char_type>
        std::ostream & operator<<( 
        std::ostream & out, 
        const boost::dbgutil::dbg_wrapper< handle_traits, wrapper_type, other_char_type> & val) {
    val.do_write( out);
    return out;
}

template< 
            class handle_traits, 
            boost::dbgutil::dbg_wrapper_type wrapper_type, 
            class other_char_type>
std::istream & operator>>( 
        std::istream & in, 
        boost::dbgutil::dbg_wrapper< handle_traits, wrapper_type, other_char_type> & val) {
    val.do_read( in);
    return in;
}

#endif


#endif // !defined(AFX_TIME_T_WRAPPER_H__5ECB1ADD_9D17_41A3_8E21_5DF960B7646A__INCLUDED_)
